/*  This file is part of distributed Structured Prediction (dSP) - http://www.alexander-schwing.de/
*
*  distributed Structured Prediction (dSP) is free software: you can
*  redistribute it and/or modify it under the terms of the GNU General
*  Public License as published by the Free Software Foundation, either
*  version 3 of the License, or (at your option) any later version.
*
*  distributed Structured Prediction (dSP) is distributed in the hope
*  that it will be useful, but WITHOUT ANY WARRANTY; without even the
*  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
*  PURPOSE. See the GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with distributed Structured Prediction (dSP).
*  If not, see <http://www.gnu.org/licenses/>.
*
*  Copyright (C) 2010-2013  Alexander G. Schwing  [http://www.alexander-schwing.de/]
*/

//Author: Alexander G. Schwing

/*  ON A PERSONAL NOTE: I spent a significant amount of time to go through
*  both, theoretical justifications and coding of this framework.
*  I hope the package is useful for your task. Any requests,
*  feedback, donations and support would be greatly appreciated.
*  Thank you for contacting me!
*/

#include "libBP.h"

#include <cmath>
#include <iostream>
#include <string.h>

#include "../Test_libHCRF/GeneralDefinitions.h"

template <class T>
libBP<T>::libBP() {
	epsilon = T(1.0);
}

template <class T>
libBP<T>::~libBP() {

}

template <class T>
int libBP<T>::Initialize(std::map<size_t, std::pair<struct PotentialStruct<T>, std::set< size_t > > > &LocalPot, std::map<size_t, std::pair<struct PotentialStruct<T>, std::set< size_t > > > &FactorPot, T epsilon) {
	this->epsilon = epsilon;
	
	num_nodes = LocalPot.size();
	num_factors = FactorPot.size();

	this->LocalPot = &LocalPot;
	this->FactorPot = &FactorPot;

	for(typename std::map<size_t, std::pair<struct PotentialStruct<T>, std::set< size_t > > >::iterator v=LocalPot.begin();v!=LocalPot.end();++v) {
		for(std::set< size_t >::iterator beta=v->second.second.begin();beta!=v->second.second.end();++beta) {
			log_n[v->first][*beta].assign(v->second.first.num_states[0], T(0.0));
		}
	}

	return 0;
}

template <class T>
T libBP<T>::computeM(size_t v, size_t alpha, size_t x_i, size_t v_num_states) {
	std::pair<struct PotentialStruct<T>, std::set<size_t> >* FactorPotAlpha = &((*FactorPot)[alpha]);

	size_t absNalpha = FactorPotAlpha->second.size();
	
	T* theta_alpha = FactorPotAlpha->first.pot;

	size_t* prodStates = new size_t[FactorPotAlpha->first.num_variables+1];
	prodStates[0] = 1;
	for(size_t k=0;k<FactorPotAlpha->first.num_variables;++k) {
		prodStates[k+1] = prodStates[k]*FactorPotAlpha->first.num_states[k];
	}

	int v_local_IX = -1;
	int local_IX = 0;
	size_t threshold = 1;
	std::set<size_t>::const_iterator endPtr_u = FactorPotAlpha->second.end();
	for(std::set<size_t>::const_iterator u=FactorPotAlpha->second.begin();u!=endPtr_u;++u, ++local_IX) {
		if(*u==v) {
			v_local_IX = local_IX;
		} else {
			threshold *= FactorPotAlpha->first.num_states[local_IX];
		}
			
	}	

	T maxVal = T(0.0);

	size_t* vars = new size_t[absNalpha];
	vars[v_local_IX] = x_i;

	T* mem = new T[threshold];

	for(size_t x_alphaNOi=0;x_alphaNOi<threshold;++x_alphaNOi) {
		size_t potentIX = 0;
		for(int k=0;k<int(absNalpha);++k) {
			if(k<v_local_IX) {
				vars[k] = (x_alphaNOi/prodStates[k])%(FactorPotAlpha->first.num_states[k]);
			} else if(k>v_local_IX) {
				vars[k] = ((x_alphaNOi*v_num_states)/prodStates[k])%(FactorPotAlpha->first.num_states[k]);
			}
			potentIX += vars[k]*prodStates[k];
		}
		
		mem[x_alphaNOi] = theta_alpha[potentIX];

		size_t varIX = 0;
		for(std::set<size_t>::const_iterator u=FactorPotAlpha->second.begin();u!=endPtr_u;++u, ++varIX) {
			if( *u!=v ) {
				mem[x_alphaNOi] += log_n[*u][alpha][vars[varIX]];
			}
		}
		maxVal = (mem[x_alphaNOi]>maxVal || x_alphaNOi==0) ? mem[x_alphaNOi] : maxVal;
	}
	
	if(epsilon!=T(0.0)) {
		T sumVal = exp( (mem[0] - maxVal)/epsilon );
		for(unsigned int x_alphaNOi=1;x_alphaNOi<threshold;++x_alphaNOi) {
			sumVal += exp( (mem[x_alphaNOi] - maxVal)/epsilon );
		}
		maxVal += epsilon*log(sumVal);
	}

	delete [] vars;
	delete [] mem;
	delete [] prodStates;

	return maxVal;
}

template <class T>
void libBP<T>::computeN() {
	typename std::map<size_t, std::pair<struct PotentialStruct<T>, std::set<size_t> > >::const_iterator endPtr_v = LocalPot->end();
	for(typename std::map<size_t, std::pair<struct PotentialStruct<T>, std::set<size_t> > >::const_iterator v=LocalPot->begin();v!=endPtr_v;++v) {
		size_t absNi = v->second.second.size();
		T* theta_i = v->second.first.pot;
		size_t num_states = v->second.first.num_states[0];

		T* sum_log_n_x_i = new T[absNi];
		T* m_alphaTOi = new T[absNi];

		memset((char*)sum_log_n_x_i, 0, sizeof(T)*absNi);

		std::map<size_t, std::vector<T> >* log_n_v = &log_n[v->first];

		for(size_t x_i=0;x_i<num_states;++x_i) {
			T tmp = (theta_i==NULL) ? T(0.0) : theta_i[x_i];

			size_t beta_local_IX = 0;
			std::set<size_t>::const_iterator endPtr_beta = v->second.second.end();
			for(std::set<size_t>::const_iterator beta = v->second.second.begin();beta!=endPtr_beta;++beta, ++beta_local_IX) {		

				m_alphaTOi[beta_local_IX] = computeM(v->first, *beta, x_i, num_states);
				tmp += m_alphaTOi[beta_local_IX];
			}

			size_t alpha_local_IX = 0;
			std::set<size_t>::const_iterator endPtr_alpha = v->second.second.end();
			for(std::set<size_t>::const_iterator alpha=v->second.second.begin();alpha!=endPtr_alpha;++alpha, ++alpha_local_IX) {
				T buf = tmp/(T(1.0) + T(absNi)) - m_alphaTOi[alpha_local_IX];
				(*log_n_v)[*alpha][x_i] = buf;
				sum_log_n_x_i[alpha_local_IX] += buf;

				/*if(x_i==0) {
					sum_log_n_x_i[alpha_local_IX] = tmp/(T(1.0) + T(absNi)) - m_alphaTOi[alpha_local_IX];
				} else {
					log_n[v->first][*alpha][x_i] = tmp/(T(1.0) + T(absNi)) - m_alphaTOi[alpha_local_IX] - sum_log_n_x_i[alpha_local_IX];
				}*/
			}
		}

		for(size_t k=0;k<absNi;++k) {
			sum_log_n_x_i[k] /= num_states;
		}

		size_t alpha_local_IX = 0;
		std::set<size_t>::const_iterator endPtr_alpha = v->second.second.end();
		for(std::set<size_t>::const_iterator alpha=v->second.second.begin();alpha!=endPtr_alpha;++alpha, ++alpha_local_IX) {
			for(size_t x_i=0;x_i<num_states;++x_i) {
				(*log_n_v)[*alpha][x_i] -= sum_log_n_x_i[alpha_local_IX];
			}
		}

		delete [] sum_log_n_x_i;
		delete [] m_alphaTOi;
	}
}

template <class T>
T libBP<T>::GetEntropySum() {
	T entropySum = T(0.0);

	if(epsilon>T(0.0)) {
		std::map<size_t, std::vector<T> > localBelief;	
		computeLocalBelief(localBelief);

		typename std::map<size_t, std::pair<struct PotentialStruct<T>, std::set<size_t> > >::const_iterator endPtr_v = LocalPot->end();
		for(typename std::map<size_t, std::pair<struct PotentialStruct<T>, std::set<size_t> > >::const_iterator v=LocalPot->begin();v!=endPtr_v;++v) {
			size_t num_states = v->second.first.num_states[0];
			T tmp = T(0.0);
			T* localBelief_v = &localBelief[v->first][0];
		
			for(size_t x_v=0;x_v<num_states;++x_v) {
				tmp -= localBelief_v[x_v]*std::log(localBelief_v[x_v]+T(1e-20));	
			}
			entropySum += epsilon*tmp;
		}

		typename std::map<size_t, std::pair<struct PotentialStruct<T>, std::set<size_t> > >::const_iterator endPtr_alpha = FactorPot->end();
		for(typename std::map<size_t, std::pair<struct PotentialStruct<T>, std::set<size_t> > >::const_iterator alpha=FactorPot->begin();alpha!=endPtr_alpha;++alpha) {
			size_t absNalpha = alpha->second.second.size();
			size_t numVars = alpha->second.first.num_variables;
			size_t* vars = new size_t[absNalpha];
			size_t* prodStates = new size_t[numVars+1];

			T tmp = T(0.0);
			for(size_t x_alpha=0;x_alpha<prodStates[numVars];++x_alpha) {
				size_t potentIX = 0;
				for(size_t k=0;k<absNalpha;++k) {
					vars[k] = (x_alpha/prodStates[k])%(alpha->second.first.num_states[k]);				
					potentIX += vars[k]*prodStates[k];
				}

				T buf = T(1.0);
				size_t neighborIX = 0;
				std::set<size_t>::const_iterator endPtr_u = alpha->second.second.end();
				for(std::set<size_t>::const_iterator u=alpha->second.second.begin();u!=endPtr_u;++u, ++neighborIX) {				
					buf *= localBelief[*u][vars[neighborIX]];
				}
				tmp -= buf*std::log(buf+T(1e-20));
			}

			entropySum += epsilon*tmp;
			delete [] vars;
			delete [] prodStates;
		}
	}

	return entropySum;
}

template <class T>
T libBP<T>::computePrimal() {		//currently only correct for epsilon = 0; otherwise we have to check agreement between local and factor potentials (the current primal is a lower bound)!!!
	std::map<size_t, std::vector<T> > localBelief;	
	computeLocalBelief(localBelief);

	T primalCost = T(0.0);

	typename std::map<size_t, std::pair<struct PotentialStruct<T>, std::set<size_t> > >::const_iterator endPtr_v = LocalPot->end();
	for(typename std::map<size_t, std::pair<struct PotentialStruct<T>, std::set<size_t> > >::const_iterator v=LocalPot->begin();v!=endPtr_v;++v) {
		T* theta_v = v->second.first.pot;
		T* localBelief_v = &localBelief[v->first][0];
		size_t num_states = v->second.first.num_states[0];

		T tmp = T(0.0);
		if(theta_v!=NULL) {
			for(size_t x_v=0;x_v<num_states;++x_v) {
				tmp += theta_v[x_v]*localBelief_v[x_v];	
			}
		}
		primalCost += tmp;

		tmp = T(0.0);
		if(epsilon>T(0.0)) {
			for(size_t x_v=0;x_v<num_states;++x_v) {
				tmp -= localBelief_v[x_v]*std::log(localBelief_v[x_v]+T(1e-20));	
			}
			primalCost += epsilon*tmp;
		}
	}

	typename std::map<size_t, std::pair<struct PotentialStruct<T>, std::set<size_t> > >::const_iterator endPtr_alpha = FactorPot->end();
	for(typename std::map<size_t, std::pair<struct PotentialStruct<T>, std::set<size_t> > >::const_iterator alpha=FactorPot->begin();alpha!=endPtr_alpha;++alpha) {
		size_t absNalpha = alpha->second.second.size();
		T* theta_alpha = alpha->second.first.pot;

		size_t numVars = alpha->second.first.num_variables;
		size_t* vars = new size_t[absNalpha];

		size_t* prodStates = new size_t[numVars+1];
		prodStates[0] = 1;
		for(size_t k=0;k<numVars;++k) {
			prodStates[k+1] = prodStates[k]*alpha->second.first.num_states[k];
		}

		T tmp = T(0.0);
		for(size_t x_alpha=0;x_alpha<prodStates[numVars];++x_alpha) {
			size_t potentIX = 0;
			for(size_t k=0;k<absNalpha;++k) {
				vars[k] = (x_alpha/prodStates[k])%(alpha->second.first.num_states[k]);				
				potentIX += vars[k]*prodStates[k];
			}

			T buf = theta_alpha[potentIX];
			size_t neighborIX = 0;
	
			std::set<size_t>::const_iterator endPtr_u = alpha->second.second.end();
			for(std::set<size_t>::const_iterator u=alpha->second.second.begin();u!=endPtr_u;++u, ++neighborIX) {				
				buf *= localBelief[*u][vars[neighborIX]];
			}
			tmp += buf;
		}

		primalCost += tmp;

		if(epsilon>T(0.0)) {
			tmp = T(0.0);
			for(size_t x_alpha=0;x_alpha<prodStates[numVars];++x_alpha) {
				size_t potentIX = 0;
				for(size_t k=0;k<absNalpha;++k) {
					vars[k] = (x_alpha/prodStates[k])%(alpha->second.first.num_states[k]);				
					potentIX += vars[k]*prodStates[k];
				}

				T buf = T(1.0);
				size_t neighborIX = 0;
				std::set<size_t>::const_iterator endPtr_u = alpha->second.second.end();
				for(std::set<size_t>::const_iterator u=alpha->second.second.begin();u!=endPtr_u;++u, ++neighborIX) {				
					buf *= localBelief[*u][vars[neighborIX]];
				}
				tmp -= buf*std::log(buf+T(1e-20));
			}

			primalCost += epsilon*tmp;
		}

		delete [] vars;
		delete [] prodStates;
	}

	return primalCost;
}

template <class T>
T libBP<T>::computeDual() {
	T dualCost = T(0.0);

	typename std::map<size_t, std::pair<struct PotentialStruct<T>, std::set<size_t> > >::const_iterator endPtr_v = LocalPot->end();
	for(typename std::map<size_t, std::pair<struct PotentialStruct<T>, std::set<size_t> > >::const_iterator v=LocalPot->begin();v!=endPtr_v;++v) {
		//size_t absNi = v->second.second.size();
		T* theta_i = v->second.first.pot;
		size_t num_states = v->second.first.num_states[0];

		T* vals = new T[num_states];
		std::map<size_t, std::vector<T> >* log_n_v = &log_n[v->first];

		T maxVal = 0;
		for(size_t x_i=0;x_i<num_states;++x_i) {
			vals[x_i] = (theta_i==NULL) ? T(0.0) : theta_i[x_i];

			std::set<size_t>::const_iterator endPtr_alpha = v->second.second.end();
			for(std::set<size_t>::const_iterator alpha=v->second.second.begin();alpha!=endPtr_alpha;++alpha) {
				vals[x_i] -= (*log_n_v)[*alpha][x_i];
			}

			maxVal = (x_i==0 || vals[x_i]>maxVal) ? vals[x_i] : maxVal;
		}
		
		dualCost += maxVal;
		
		if(epsilon>0) {
			T sum = T(0.0);
			for(size_t x_i=0;x_i<num_states;++x_i) {
				sum += exp( (vals[x_i] - maxVal)/epsilon );
			}
			dualCost += epsilon*std::log(sum+T(1e-20));
		}

		delete [] vals;
	}

	typename std::map<size_t, std::pair<struct PotentialStruct<T>, std::set<size_t> > >::const_iterator endPtr_alpha = FactorPot->end();
	for(typename std::map<size_t, std::pair<struct PotentialStruct<T>, std::set<size_t> > >::const_iterator alpha=FactorPot->begin();alpha!=endPtr_alpha;++alpha) {
		size_t absNalpha = alpha->second.second.size();
		T* theta_alpha = alpha->second.first.pot;

		size_t numVars = alpha->second.first.num_variables;
		size_t* vars = new size_t[absNalpha];

		size_t* prodStates = new size_t[numVars+1];
		prodStates[0] = 1;
		for(size_t k=0;k<numVars;++k) {
			prodStates[k+1] = prodStates[k]*alpha->second.first.num_states[k];
		}

		T* vals = new T[prodStates[numVars]];		

		T maxVal = 0;
		for(size_t x_alpha=0;x_alpha<prodStates[numVars];++x_alpha) {
			size_t potentIX = 0;
			for(size_t k=0;k<absNalpha;++k) {
				vars[k] = (x_alpha/prodStates[k])%(alpha->second.first.num_states[k]);				
				potentIX += vars[k]*prodStates[k];
			}
			
			vals[x_alpha] = theta_alpha[potentIX];

			size_t varIX = 0;
			std::set<size_t>::const_iterator endPtr_u = alpha->second.second.end();
			for(std::set<size_t>::const_iterator u=alpha->second.second.begin();u!=endPtr_u;++u, ++varIX) {
				vals[x_alpha] += log_n[*u][alpha->first][vars[varIX]];
			}

			maxVal = (x_alpha==0 || vals[x_alpha]>maxVal) ? vals[x_alpha] : maxVal;
		}

		dualCost += maxVal;

		if(epsilon>0) {
			T sum = T(0.0);
			for(size_t x_alpha=0;x_alpha<prodStates[numVars];++x_alpha) {
				sum += exp( (vals[x_alpha] - maxVal)/epsilon );
			}
			dualCost += epsilon*std::log(sum+T(1e-20));
		}

		delete [] vars;
		delete [] vals;
		delete [] prodStates;
	}

	return dualCost;
}

template <class T>
void libBP<T>::computeLocalBelief(std::map<size_t, std::vector<T> > &localBelief) {
	typename std::map<size_t, std::pair<struct PotentialStruct<T>, std::set<size_t> > >::const_iterator endPtr_v = LocalPot->end();
	for(typename std::map<size_t, std::pair<struct PotentialStruct<T>, std::set<size_t> > >::const_iterator v=LocalPot->begin();v!=endPtr_v;++v) {
		//size_t absNi = v->second.second.size();
		T* theta_i = v->second.first.pot;
		size_t num_states = v->second.first.num_states[0];
		localBelief[v->first].assign(num_states, T(0.0));

		T* localBelief_i = &localBelief[v->first][0];
		std::map<size_t, std::vector<T> >* log_n_v = &log_n[v->first];

		T maxVal = T(0.0);
		for(size_t x_i=0;x_i<num_states;++x_i) {
			localBelief_i[x_i] = (theta_i==NULL) ? T(0.0) : theta_i[x_i];

			std::set<size_t>::const_iterator endPtr_alpha = v->second.second.end();
			for(std::set<size_t>::const_iterator alpha=v->second.second.begin();alpha!=endPtr_alpha;++alpha) {
				localBelief_i[x_i] -= (*log_n_v)[*alpha][x_i];
			}

			maxVal = (localBelief_i[x_i]>maxVal || x_i==0) ? localBelief_i[x_i] : maxVal;
		}
		T sum = T(0.0);
		for(size_t x_i=0;x_i<num_states;++x_i) {
			if(epsilon==0) {
				localBelief_i[x_i] = (localBelief_i[x_i] - maxVal)>-1e-5 ? T(1.0) : T(0.0);
			} else {
				localBelief_i[x_i] = (exp( (localBelief_i[x_i] - maxVal)/epsilon ) );
			}
			sum += localBelief_i[x_i];
		}
		for(size_t x_i=0;x_i<num_states;++x_i) {
			localBelief_i[x_i] /= sum;
		}
	}
}

template <class T>
int libBP<T>::RunBP(int MaxIter, bool debug) {

	//std::vector<T> DualValues(MaxIter, T(0.0));
	//std::vector<T> PrimalValues(MaxIter, T(0.0));

	CPrecisionTimer CTmr;
	if(debug) {
		CTmr.Start();
	}

	for(int iteration=0;iteration<MaxIter;++iteration) {
		computeN();

		if(debug) {
			T dual = computeDual();
			T primal = computePrimal();
			//std::cout << "Dual: " << dual << "\tPrimal: " << primal << std::endl;

			if(dual==primal) {
				double time = CTmr.Stop();
				std::cout << iteration << "\tTime: " << time << "\tDual: " << dual << "\tPrimal: " << primal << "\tGap: " << dual - primal << std::endl;
				break;
			}
			if(iteration==MaxIter-1) {
				double time = CTmr.Stop();
				std::cout << iteration << "\tTime: " << time << "\tDual: " << dual << "\tPrimal: " << primal << "\tGap: " << dual - primal << std::endl;
			}
		}
		//DualValues[iteration] = computeDual();
		//PrimalValues[iteration] = computePrimal();
		//std::cout << DualValues[iteration] << " " << curTime << " " << PrimalValues[iteration] << std::endl;
	}

	return 0;

}

template <class T>
int libBP<T>::GetResult(std::map<size_t, std::vector<T> > &localBelief) {	
	computeLocalBelief(localBelief);
	return 0;
}

template class libBP<double>;
template class libBP<float>;